'use strict';

const _ = require('lodash');
const hostFunctions = require('./lib/host');

/**
 * Apache functions.
 * @namespace servers.apache
 */

$app.helpers.setupBitnamiConfig = function() {
  $file.mkdir($app.vhostsDir);
  $file.mkdir($app.htaccessDir);
  // Write basic Bitnami customizations
  $file.append($app.confFile, `PidFile "${$app.pidFile}"
TraceEnable Off
Include ${$app.confDir}/deflate.conf
IncludeOptional ${$app.vhostsDir}/*.conf
Include ${$app.confDir}/bitnami/bitnami.conf`, {atNewLine: true});
  // Enable main virtual hosts
  $hb.renderToFile('bitnami.conf.tpl', `${$app.confDir}/bitnami/bitnami.conf`);
  $hb.renderToFile('bitnami-ssl.conf.tpl', `${$app.confDir}/bitnami/bitnami-ssl.conf`);
  // Enable server-status virtual hosts
  $hb.renderToFile('status-vhost.conf.tpl', $file.join($app.confDir, 'vhosts/00_status-vhost.conf'));
};

$app.helpers.httpoxyPatch = function() {
  $app.info('==> Patching httpoxy...');
  const unsetProxyText = `<IfModule mod_headers.c>
   RequestHeader unset Proxy
</IfModule>`;
  $file.append($app.confFile, unsetProxyText, {atNewLine: true});
};

$app.helpers.setupPorts = function() {
  const options = {abortOnUnmatch: true, type: 'regexp'};
  const listenPattern = /^\s*Listen\s+([^:]*:)?[0-9]+\s*$/gm;
  const serverNamePattern = /^\s*#?\s*ServerName\s+([^:\s]+)(:[0-9]+)?$/gm;
  const vhostPattern = /VirtualHost\s+([^:>]+)(:[0-9]+)/gm;

  if ($app.enableCustomPorts === 'yes') {
    $file.substitute($app.confFile, [{
      pattern: listenPattern,
      value: `Listen ${$app.httpPort}`,
    }, {
      pattern: serverNamePattern,
      value: `ServerName $1:${$app.httpPort}`,
    }], options);
    $file.substitute($file.join($app.confDir, 'vhosts/00_status-vhost.conf'), [{
      pattern: vhostPattern,
      value: `VirtualHost $1:${$app.httpPort}`,
    }], options);
    $file.substitute($file.join($app.confDir, 'bitnami/bitnami.conf'), [{
      pattern: vhostPattern,
      value: `VirtualHost $1:${$app.httpPort}`,
    }], options);
    $file.substitute($file.join($app.confDir, 'bitnami/bitnami-ssl.conf'), [{
      pattern: listenPattern,
      value: `Listen ${$app.httpsPort}`,
    }, {
      pattern: vhostPattern,
      value: `VirtualHost $1:${$app.httpsPort}`,
    }]);
  }
};

/**
 * Change configuration to adapt it to host memory
 * @function servers.apache~resize
 * @param {Object} [options]
 * @param {Object} [options.memory] - Max memory to use. Uses maximum available by default (Supports M or G)
 * @example
 * resize({memory: '2G'});
 */
$app.helpers.resize = function(options) {
  const sizes = {
    'micro': {
      prefork: {
        StartServers: 5, MinSpareServers: 5, MaxSpareServers: 10, MaxRequestWorkers: 5, KeepAliveTimeout: 1,
      },
      event: {
        ServerLimit: 4, StartServers: 2, MinSpareThreads: 128, MaxSpareThreads: 192, MaxRequestWorkers: 256,
      },
    },
    'small': {
      prefork: {
        StartServers: 5, MinSpareServers: 5, MaxSpareServers: 10, MaxRequestWorkers: 15, KeepAliveTimeout: 2,
      },
      event: {
        ServerLimit: 8, StartServers: 3, MinSpareThreads: 192, MaxSpareThreads: 384, MaxRequestWorkers: 512,
      },
    },
    'medium': {
      prefork: {
        StartServers: 5, MinSpareServers: 5, MaxSpareServers: 15, MaxRequestWorkers: 30, KeepAliveTimeout: 5,
      },
      event: {
        ServerLimit: 16, StartServers: 4, MinSpareThreads: 384, MaxSpareThreads: 640, MaxRequestWorkers: 1024,
      },
    },
    'large': {
      prefork: {
        StartServers: 10, MinSpareServers: 10, MaxSpareServers: 30, MaxRequestWorkers: 150, KeepAliveTimeout: 5,
      },
      event: {
        ServerLimit: 24, StartServers: 8, MinSpareThreads: 512, MaxSpareThreads: 3768, MaxRequestWorkers: 1536,
      },
    },
    'xlarge': {
      prefork: {
        StartServers: 10, MinSpareServers: 10, MaxSpareServers: 30, MaxRequestWorkers: 150, KeepAliveTimeout: 5,
      },
      event: {
        ServerLimit: 24, StartServers: 8, MinSpareThreads: 512, MaxSpareThreads: 768, MaxRequestWorkers: 1536,
      },
    },
    '2xlarge': {
      prefork: {
        StartServers: 10, MinSpareServers: 10, MaxSpareServers: 30, MaxRequestWorkers: 150, KeepAliveTimeout: 5,
      },
      event: {
        ServerLimit: 24, StartServers: 8, MinSpareThreads: 512, MaxSpareThreads: 768, MaxRequestWorkers: 1536,
      },
    },
  };
  const machineSize = hostFunctions.getMachineSize(options);
  $file.mkdir($file.join($app.confDir, 'bitnami'));
  const memoryConfFile = $file.join($app.confDir, 'bitnami/memory.conf');
  $hb.renderToFile('memory.conf.tpl', memoryConfFile, sizes[machineSize]);
  if (!$file.contains($app.confFile, `Include "${memoryConfFile}"`)) {
    $file.puts($app.confFile, `# Memory settings\nInclude "${memoryConfFile}"`, {atNewLine: true});
  }
};

/**
 * Enable or disable the specified apache module
 * @function servers.apache~configureModules
 * @param {Array} [modules] - Array of [string] moduleName and/or [Object] {name: moduleName, id: moduleID}
 *                            (regular expresions are allowed for moduleName strings)
 * @param {string} [state] - Module state [enable/disable]
 * @param {Object} [options] - Options object
 * @param {boolean} [options.restartIfRunning=false] - Restart webserver after set module (required to take effect)
 * @example
 * // Enable rewrite_module -> LoadModule rewrite_module modules/mod_rewrite.so
 * configureModules(['mod_rewrite'], 'enable');
 * @example
 * // Enable php5_module -> LoadModule php5_module modules/libphp5.so
 * configureModules([{name: 'libphp5', id: 'php5_module'}], 'enable');
 * @example
 * // Enable php5_module and restart Apache webserver -> LoadModule php5_module modules/libphp5.so
 * configureModules([{name: 'libphp5', id: 'php5_module'}], 'enable', {restartIfRunning: true});
 * @example
 * // Disable rewrite_module -> #LoadModule rewrite_module modules/mod_rewrite.so
 * configureModules(['mod_rewrite'], 'disable');
 */
$app.helpers.configureModules = function(modules, action, options) {
  options = _.defaults(options || {}, {restartIfRunning: false});
  if (!_.isArray(modules)) {
    throw new Error('configureModules function requires an array of modules as first parameter');
  }
  let matched = false;
  const confFile = $file.join($app.confDir, 'httpd.conf');
  modules.forEach(function(module) {
    let moduleName = module;
    let moduleID = '';
    if (_.isObject(module)) {
      if (module.name && module.id) {
        moduleName = module.name;
        moduleID = module.id;
      } else {
        throw new Error(`The object ${module} should have the fields module.name and module.id`);
      }
    }
    if (_.isEqual(action, 'enable')) {
      try {
        // Looking for "#LoadModule <whatever> <module>.so" to match the exact module name
        $file.substitute(
          confFile,
          new RegExp(`^\\s*#*\\s*(LoadModule\\s*[^\\s]*\\s*modules/${moduleName}\\.so.*)$`, 'm'),
          '$1',
          {abortOnUnmatch: true}
        );
      } catch (err) {
        // Module is not present in the file, write it at the end
        if (!_.isEmpty(moduleID)) {
          $file.eachLine(confFile, function(line) {
            if (!matched && line.match(/^\s*LoadModule/)) {
              matched = true;
              return `${line}\nLoadModule ${moduleID} modules/${moduleName}.so`;
            }
            return line;
          }, function(text) {
            if (matched) {
              $file.write(confFile, text);
            }
          }, {reverse: true});
        } else {
          throw new Error(`Module ${moduleName} is not present in the configuration file, `
            + 'you need to provide a module identifier in the options parameter along with the module name');
        }
      }
    } else if (_.isEqual(action, 'disable')) {
      // Disabling if exists
      $file.substitute(confFile, [{
        pattern: new RegExp(`^\\s*(LoadModule\\s*[^\\s]*\\s*modules/${moduleName}\\.so.*)$`, 'm'),
        value: '#$1',
      }]);
    } else {
      throw new Error(`Action '${action}' not supported`);
    }
  });
  if (options.restartIfRunning === true && $app.status.isRunning === 'true') {
    $app.helpers.restart();
  }
};

/**
 * Collect all the .htaccess files from /opt/bitnami/`name`
 * and write the result (`name`-htaccess.conf) in the htaccess directory
 * @function servers.apache~replaceHtaccessFiles
 * @param {string} name - Application name
 * @param {Object} [options] - Options object
 * @param {boolean} [options.overwrite=true] - Overwrite the original .htaccess with explanation text
 * @return {string} - Return the resulted file
 * @example
 * // Collect all the magento .htaccess
 * replaceHtaccessFiles('magento');
 * // => /opt/bitnami/apache/conf/vhosts/htaccess/magento-htaccess.conf contains the result
 */
$app.helpers.replaceHtaccessFiles = function(name, options) {
  options = _.defaults(options || {}, {overwrite: true});
  const resultFile = $file.join($app.htaccessDir, `${name}-htaccess.conf`);
  const htaccessList = $file.glob(
    '**/.htaccess',
    {
      cwd: $file.join($file.dirname($app.installdir), name),
      absolutize: true,
    }
  );
  _.forEach(htaccessList, (htaccessFile) => {
    const data = $file.read(htaccessFile);
    if (!$file.contains(htaccessFile, '# This configuration has been moved to')) {
      $file.puts(resultFile, `<Directory "${$file.dirname(htaccessFile)}">
${data}
</Directory>`);
    }
    if (options.overwrite) {
      $file.write(htaccessFile, `# This configuration has been moved to ${resultFile} `
                  + 'config file for performance and security reasons');
    }
    $file.chmod(htaccessFile, '640');
  });
};

/**
 * Delete the `name`-htaccess.conf
 * @function servers.apache~removeHtaccessFile
 * @param {string} name - Application name for which remove the htaccess configuration file
 * @example
 * // Delete htaccess configuration file
 * removeHtaccessFile($app.name);
 */
$app.helpers.removeHtaccessFile = function(name) {
  $file.delete($file.join($app.htaccessDir, `${name}-htaccess.conf`));
};

/**
 * Add an application virtual host
 * @function servers.apache~addVhost
 * @param {string} name - Application name for which configure the virtual host
 * @param {Object} [options] - Options object
 * @param {string} [options.host=_default_] - Server host
 * @param {number} [options.port=80] - Port to use
 * @param {string} [options.documentRoot] - Application document root
 * @param {string} [options.requireOption] - Application Require option
 * @param {string} [options.proxyUrl] - Url to proxy requests
 * @param {boolean} [options.enableHttps=false] - Create another virtual Host using HTTPS certificates
 * @param {string} [options.certificatePath='conf/bitnami/certs/'] - Path to certificates server.crt and server.key
 * @param {string} [options.vhostText] - Use given text to generate the vhost
 * @example
 * // Configure default vhost file
 * addVhost($app.name);
 */
$app.helpers.addVhost = function(name, options) {
  options = _.defaults(options || {}, {
    host: '_default_',
    port: 80,
    documentRoot: '',
    requireOption: 'all granted',
    customConf: '',
    proxyUrl: '',
    enableHttps: false,
    certificatePath: $file.join($app.confDir, 'bitnami/certs/'),
    vhostText: '',
  });
  if (!_.isEmpty(options.vhostText)) {
    $file.write($file.join($app.vhostsDir, `${name}-vhost.conf`), options.vhostText);
  } else {
    if (_.isEmpty(options.documentRoot) && _.isEmpty(options.proxyUrl) && _.isEmpty(options.customConf)) {
      throw new Error('You should either specify a DocumentRoot, a ProxyPass URL or a custom configuration');
    }
    const renderOptions = {
      host: options.host,
      port: options.port,
      documentRoot: options.documentRoot ? `  DocumentRoot ${options.documentRoot}
  <Directory "${options.documentRoot}">
    Options -Indexes -FollowSymLinks +MultiViews
    AllowOverride All
    Require ${options.requireOption}
  </Directory>\n` : '',
      sslInfo: options.enableHttps ? `  SSLEngine on
  SSLCertificateFile "${options.certificatePath}/server.crt"
  SSLCertificateKeyFile "${options.certificatePath}/server.key"` : '',
      proxyPass: options.proxyUrl ? `  ProxyPass ${options.proxyUrl}
  ProxyPassReverse ${options.proxyUrl}` : '',
      customConf: options.customConf,
    };
    $hb.renderToFile('httpd-vhost.conf.tpl', $file.join($app.vhostsDir, `${name}-vhost.conf`), renderOptions);
  }
};

/**
 * Delete the name-vhost.conf
 * @function servers.apache~removeVhost
 * @param {string} name - Application name for which configure the virtual host
 * @example
 * // Delete vhost file
 * removeVhost($app.name);
 */
$app.helpers.removeVhost = function(name) {
  $file.delete($file.join($app.vhostsDir, `${name}-vhost.conf`));
};

/**
 * Add or remove configuration lines to httpd.conf file
 * @function servers.apache~configurationEntry
 * @param {string} entry - Configuration line we want to add
 * @param {string} action - Action to perform [add/remove/enable/disable]
 * @param {Object} [options] - Options object
 * @param {boolean} [options.reloadIfRunning=false] - Reload configuration after configure
 * @example
 * // Add entry
 * configureEntry('Include vhosts.conf', 'add');
 * // Disable entry and then reload Apache configuration
 * configureEntry('Include vhosts.conf', 'disable', {reloadIfRunning: true})
 */
$app.helpers.configureEntry = function(entry, action, options) {
  options = _.defaults(options || {}, {reloadIfRunning: false});
  const confFile = $file.join($app.confDir, 'httpd.conf');
  switch (action) {
    case 'add':
      // Add if doesn't match
      if (!$file.read(confFile).match(new RegExp(`^\\s*${entry}\\s*$`, 'm'))) {
        $file.puts(confFile, entry, {atNewLine: true});
      }
      break;
    case 'remove':
      $file.substitute(confFile, new RegExp(`^\\s*(#?)(${entry}\\s*)$`, 'm'), '');
      break;
    case 'enable':
      $file.substitute(confFile, new RegExp(`^\\s*[#]*\\s*(${entry}\\s*)$`, 'm'), '$1');
      break;
    case 'disable':
      $file.substitute(confFile, new RegExp(`^\\s*(#?)(${entry}\\s*)$`, 'm'), '#$2');
      break;
    default:
      throw new Error(`Action '${action} not supported`);
  }
  if (options.reloadIfRunning && $app.status().isRunning) {
    $app.helpers.reload();
  }
};

/**
 * Reload the apache configuration without restarting
 * @function servers.apache~reload
 * @example
 * reload();
 */
$app.helpers.reload = function() {
  if ($app.status().isRunning) {
    $os.runProgram($file.join($app.binDir, 'apachectl'), 'graceful');
  } else {
    $app.trace(`Skipping reload: ${$app.status().statusOutput}`);
  }
};
